
open Printf

(** Termes *)

type op = Add | Sub | Mul | Div

let char_of_op = function
  | Add -> '+' | Sub -> '-'
  | Mul -> '*' | Div -> '/'

type term =
  | Fun of string * term
  | App of term * term
  | Var of string
  | Num of int
  | Ifz of term * term * term
  | Op  of op * term * term
  | Let of string * term * term
  | FixFun of string * string * term

let rec p_term chan = function
  | Num n -> fprintf chan "%i" n
  | Var x -> fprintf chan "%s" x
  | Op (op,t1,t2) ->
      fprintf chan "%a%c%a"
        p_arg t1
        (char_of_op op)
        p_arg t2
  | App (t1,t2) ->
      fprintf chan "%a %a" p_fun t1 p_arg t2
  | Fun (x,t1) ->
      fprintf chan "Fun %s -> %a" x p_term t1
  | FixFun (f,x,t1) ->
      fprintf chan "Fix %s -> Fun %s -> %a" f x p_term t1
  | Let (x,tx,t) ->
      fprintf chan "Let %s = %a In %a"
        x p_term tx p_term t
  | Ifz (t1,t2,t3) ->
      fprintf chan "Ifz %a Then %a Else %a"
        p_term t1 p_term t2 p_term t3

and p_arg chan t = match t with
  | Num _ | Var _ ->
      p_term chan t
  | _ -> fprintf chan "(%a)" p_term t

and p_fun chan t = match t with
  | Num _ | Var _ | App (_,_) -> p_term chan t
  | _ -> fprintf chan "(%a)" p_term t

(* Instructions *)

type instruction =
 | Ldi of int
 | Push
 | IOp of op
 | Extend
 | Search of int
 | Pushenv
 | Popenv
 | Mkclos of instruction list
 | Apply
 | Test of instruction list * instruction list

let string_of_instr = function
  | Ldi i    -> sprintf "Ldi %i" i
  | Push     -> "Push"
  | IOp op   -> sprintf "IOp(%c)" (char_of_op op)
  | Extend   -> "Extend"
  | Search k -> sprintf "Search %i" k
  | Pushenv  -> "Pushenv"
  | Popenv   -> "Popenv"
  | Mkclos _ -> "Mkclos"
  | Apply    -> "Apply"
  | Test _   -> "Test"

let print_instr_ln i = Printf.printf "%s\n" (string_of_instr i)
