
open Definitions
open Machine
open Compiler
open Genlex

let lexer =
  make_lexer ["Fun"; "->"; "("; ")"; "Ifz"; "Then"; 
              "Else"; "Let"; "Rec"; "="; "In"; "+"; "-"; "*"; "/"; "Fix"; ";"]

let rec parse_main = parser
  | [< e = parse_expr; 'Kwd ";"; 'Kwd ";" >] -> e

and parse_expr = parser
  | [< 'Kwd "Fun"; 'Ident x; 'Kwd "->"; e = parse_expr >] -> Fun(x,e)
  | [< 'Kwd "Fix"; 'Ident f; 'Kwd "->";
       'Kwd "Fun"; 'Ident x; 'Kwd "->"; e = parse_expr >] -> FixFun(f,x,e)
  | [< 'Kwd "Ifz"; e1 = parse_expr; 'Kwd "Then"; e2 = parse_expr; 
       'Kwd "Else"; e3 = parse_expr >] -> Ifz(e1,e2,e3)
  | [< 'Kwd "Let"; e = parse_let >] -> e
  | [< e1 = parse_par; e = parse_aux e1>] -> e

and parse_let = parser
  | [< 'Kwd "Rec"; 'Ident f; 'Kwd "="; 'Kwd "Fun"; 'Ident x; 'Kwd "->";
       e1 = parse_expr; 'Kwd "In"; e2 = parse_expr >] ->
      Let(f,FixFun(f,x,e1),e2)
  | [< 'Ident x; 'Kwd "="; e1 = parse_expr; 
       'Kwd "In"; e2 = parse_expr >] -> Let(x,e1,e2)

and parse_par = parser
  | [< 'Int n >] -> Num n
  | [< 'Ident x >] -> Var x
  | [< 'Kwd "("; e1 = parse_expr; 'Kwd ")">] -> e1

and parse_aux e1 = parser
  | [< 'Kwd "+"; e2 = parse_par >] -> Op (Add,e1,e2)
  | [< 'Kwd "-"; e2 = parse_par >] -> Op (Sub,e1,e2)
  | [< 'Kwd "*"; e2 = parse_par >] -> Op (Mul,e1,e2)
  | [< 'Kwd "/"; e2 = parse_par >] -> Op (Div,e1,e2)
  | [< e2 = parse_par_list e1 >] -> e2

and parse_par_list e = parser
  | [< e1 = parse_par; e2 = parse_par_list (App(e, e1)) >] -> e2
  | [< >] -> e

let print_term t = p_term stdout t

let () = 
  while true do
    Printf.printf "\nPCF < %!" ;
    try 
      let aux = lexer (Stream.of_channel stdin) in
      let t = parse_main aux in
      let c = Compiler.compile t in
      let s = Machine.exec c in
        Printf.printf "\n%s\n" (Machine.print s)
    with 
      | Stream.Error "" | Stream.Failure -> Printf.printf "Syntax error!\n"
      | e -> Printf.printf "Error: %s!\n" (Printexc.to_string e)
  done
